#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the Kramers-Kronig Calculator software package.
#
# Copyright (c) 2013 Benjamin Watts, Daniel J. Lauk
#
# The software is licensed under the terms of the zlib/libpng license.
# For details see LICENSE.txt

"""This module implements a GUI using the wxPython toolkit."""
import logging
logger = logging.getLogger(__name__)
if __name__ == '__main__':
	import sys
	logging.basicConfig(level=logging.DEBUG)
	logging.StreamHandler(stream=sys.stdout)

import wx
import wx.lib.plot as plot
import numpy
import os
import kk, data

try:
	import scipy.optimize
	SCIPY_FLAG = True
except ImportError:
	SCIPY_FLAG = False
	logger.info('Failed to import the scipy.optimize module - disabling the \'fix distortions\' checkbox.')


class MyFrame(wx.Frame):

	def __init__(self):
		wx.Frame.__init__(self, None, wx.ID_ANY, "Kramers-Kronig Calculator", size=(500, 800))

		# Initialise variables
		self.dirname = ''
		self.raw_file = None
		self.total_asf = None
		self.total_Im_coeffs = None
		self.merged_Im = None
		self.nexafs_CutOut = []
		self.MolecularMass = 1
		self.asf_bg = None
		#New set of variables to initialise. All those above might want to be removed.
		self.ChemicalFormula = None
		self.Stoichiometry = None
		self.Relativistic_Correction = None
		self.NearEdgeData = None
		self.splice_ind = None
		self.ASF_E = None
		self.ASF_Data = None
		self.Full_E = None
		self.Imaginary_Spectrum = None
		self.KK_Real_Spectrum = None


		# Setting up the menus.
		filemenu = wx.Menu()
		filemenu.Append(wx.ID_OPEN, "L&oad", " Load photoabsorption data from file")
		filemenu.AppendSeparator()
		filemenu.Append(wx.ID_SAVE, "&Save", " Export results to file")
		exportmenu = wx.Menu()
		exportmenu.Append(201,"Photoabsorption", " Export X-ray absorption data")
		exportmenu.Append(202,"Refractive Index", " Export beta and delta")
		filemenu.AppendSubMenu(exportmenu,"Export")   # Adding the "exportmenu" to the filemenu
		filemenu.AppendSeparator()
		filemenu.Append(wx.ID_EXIT, "E&xit", " Terminate the program")
		helpmenu = wx.Menu()
		helpmenu.Append(wx.ID_HELP, "&Help", " How to use this program")
		helpmenu.Append(205, "&Article", " Publication describing how this calculation works")
		helpmenu.AppendSeparator()
		helpmenu.Append(wx.ID_ABOUT, "&About", " Information about this program")
		# Creating the menubar.
		menuBar = wx.MenuBar()
		menuBar.Append(filemenu, "&File")  # Adding the "filemenu" to the MenuBar
		menuBar.Append(helpmenu, "&Help")  # Adding the "helpmenu" to the MenuBar
		self.SetMenuBar(menuBar)  # Adding the MenuBar to the Frame content.
		self.Bind(wx.EVT_MENU, self.OnOpen, id=wx.ID_OPEN)
		self.Bind(wx.EVT_MENU, self.OnSave, id=wx.ID_SAVE)
		self.Bind(wx.EVT_MENU, self.OnSave, id=201)   # will set convert_to="photoabsorption" when ID is recognised
		self.Bind(wx.EVT_MENU, self.OnSave, id=202)   # will set convert_to="refractive_index" when ID is recognised
		self.Bind(wx.EVT_MENU, self.OnExit, id=wx.ID_EXIT)
		self.Bind(wx.EVT_MENU, self.OnAbout, id=wx.ID_ABOUT)
		self.Bind(wx.EVT_MENU, self.OnArticle, id=205)
		self.Bind(wx.EVT_MENU, self.OnHelp, id=wx.ID_HELP)


		Sizer1 = wx.BoxSizer(wx.HORIZONTAL)  # create outer sizer
		SizerL = wx.BoxSizer(wx.VERTICAL)  # create left-hand sizer for controls
		SizerR = wx.BoxSizer(wx.VERTICAL)  # create right-hand sizer for plots
		############################Data box
		DataBox = wx.StaticBoxSizer(wx.StaticBox(self, label="Near-Edge Data"), wx.VERTICAL)
		self.FileText = wx.StaticText(self, -1, "File: (None)")
		DataBox.Add(self.FileText, 1, wx.GROW)
		DataTypeLabel = wx.StaticText(self, -1, "Data Type: ")
		self.DataTypeCombo = wx.ComboBox(self, -1, value='Photoabsorption', style=wx.CB_READONLY)
		self.DataTypeCombo.Append('Photoabsorption')
		self.DataTypeCombo.Append('Beta')
		self.DataTypeCombo.Append('Scattering Factor')
		self.DataTypeCombo.Bind(wx.EVT_COMBOBOX, self.MergeAdd_check)
		DataTypeSizer = wx.BoxSizer(wx.HORIZONTAL)
		DataTypeSizer.Add(DataTypeLabel)
		DataTypeSizer.Add(self.DataTypeCombo, 2, wx.GROW)
		DataBox.Add(DataTypeSizer, 1, wx.GROW)
		SpliceSizer = wx.BoxSizer(wx.HORIZONTAL)
		self.SpliceText1 = wx.TextCtrl(self, -1, "Start", style=wx.TE_PROCESS_ENTER)
		self.SpliceText1.Bind(wx.EVT_KILL_FOCUS, self.Splice_Text_check)
		self.SpliceText1.Bind(wx.EVT_TEXT_ENTER, self.Splice_Text_check)
		SpliceSizer.Add(self.SpliceText1, 1)
		self.SpliceText2 = wx.TextCtrl(self, -1, "End", style=wx.TE_PROCESS_ENTER)
		self.SpliceText2.Bind(wx.EVT_KILL_FOCUS, self.Splice_Text_check)
		self.SpliceText2.Bind(wx.EVT_TEXT_ENTER, self.Splice_Text_check)
		SpliceSizer.Add(self.SpliceText2, 1)
		DataBox.Add(SpliceSizer, 1, wx.GROW)

		# Background_CloseSizer = wx.BoxSizer(wx.HORIZONTAL)
#		self.InvertDataCheckBox = wx.CheckBox(self, -1, "Invert Data")
#		self.InvertDataCheckBox.Bind(wx.EVT_CHECKBOX, self.Splice_Text_check)
#		DataBox.Add(self.InvertDataCheckBox, 0)
		self.AddBackgroundCheckBox = wx.CheckBox(self, -1, "Add background")
		self.AddBackgroundCheckBox.Bind(wx.EVT_CHECKBOX, self.Splice_Text_check)
		self.AddBackgroundCheckBox.Disable()
		self.AddBackgroundCheckBox.SetToolTip(wx.ToolTip("Not implemented"))
		DataBox.Add(self.AddBackgroundCheckBox, 0)
		self.FixDistortionsCheckBox = wx.CheckBox(self, -1, "Fix distortions")
		self.FixDistortionsCheckBox.Bind(wx.EVT_CHECKBOX, self.Splice_Text_check)
		if not SCIPY_FLAG:
			self.FixDistortionsCheckBox.Disable()
			self.FixDistortionsCheckBox.SetToolTip(wx.ToolTip("Install the SciPy module to use this feature"))
		DataBox.Add(self.FixDistortionsCheckBox, 0)
		# Background_CloseSizer.Add(self.AddBackgroundCheckBox, 0)
		# self.AddBackgroundCheckBox.Bind(wx.EVT_CHECKBOX, self.MergeAdd_check)
		# Background_CloseSizer.AddStretchSpacer(1)
		# self.CloseFile = wx.Button(self, -1, "X", style= wx.BU_EXACTFIT)
		# Background_CloseSizer.Add(self.CloseFile, 0)
		# DataBox.Add(Background_CloseSizer, 1, wx.GROW)

		############################Material box
		self.MaterialBox = wx.StaticBoxSizer(wx.StaticBox(self, label="Material"), wx.VERTICAL)
		DensitySizer = wx.BoxSizer(wx.HORIZONTAL)
		DensitySizer.Add(wx.StaticText(self, -1, "Density: "))
		self.DensityText = wx.TextCtrl(self, -1, "1", style=wx.TE_PROCESS_ENTER)
		self.DensityText.Bind(wx.EVT_KILL_FOCUS, self.Splice_Text_check)
		self.DensityText.Bind(wx.EVT_TEXT_ENTER, self.Splice_Text_check)
		DensitySizer.Add(self.DensityText, 1)
		DensitySizer.Add(wx.StaticText(self, -1, " g/ml"))
		self.MaterialBox.Add(DensitySizer, 0)
		
		StoichiometrySizer = wx.BoxSizer(wx.HORIZONTAL)
		StoichiometrySizer.Add(wx.StaticText(self, -1, "Stoichiometry: "))
		self.StoichiometryText = wx.TextCtrl(self, -1, "", style=wx.TE_PROCESS_ENTER)
		self.StoichiometryText.Bind(wx.EVT_KILL_FOCUS, self.Stoichiometry_Text_check)
		self.StoichiometryText.Bind(wx.EVT_TEXT_ENTER, self.Stoichiometry_Text_check)
		StoichiometrySizer.Add(self.StoichiometryText, 1)
		self.MaterialBox.Add(StoichiometrySizer, 0)

		############################Calc box
		CalcBox = wx.StaticBoxSizer(wx.StaticBox(self, label="Calculation"), wx.VERTICAL)
		CalcButton = wx.Button(self, -1, "Calculate")
		CalcBox.Add(CalcButton, 1, wx.GROW)
		CalcButton.Bind(wx.EVT_BUTTON, self.calculate)



		SizerL.Add(DataBox, 0, wx.GROW)
		SizerL.Add(self.MaterialBox, 1, wx.GROW)
		SizerL.AddStretchSpacer(1)
		SizerL.Add(CalcBox, 0, wx.GROW)




		self.PlotAxes = plot.PlotCanvas(self)

		SizerR.Add(self.PlotAxes, 1, wx.GROW)
		#SizerR.Add(self.Rplot, 1, wx.GROW)
		# enable the zoom feature (drag a box around area of interest)
		self.PlotAxes.enableZoom = True
		#self.Rplot.SetEnableZoom(True)


		Sizer1.Add(SizerL, 1, wx.GROW)
		Sizer1.Add(SizerR, 3, wx.GROW)
		self.SetAutoLayout(True)
		self.SetSizer(Sizer1)			   # add outer sizer to frame
		self.Fit()

		self.Show(True)
		self.plot_data()
		#self.Test()

	def Test(self):
		"""Convenience function for repetitive testing"""
		self.filename = "NC-Xy_norm_bgsub.txt"
		self.dirname = "data"
		self.FileText.SetLabel("File: "+self.filename)
		#self.raw_file = self.LoadData(os.path.join(self.dirname, self.filename))
		self.AddBackgroundCheckBox.SetValue(True)
		self.combine_data()
		self.PP_AlgorithmRadio.SetValue(True)
		self.plot_data()




	def OnAbout(self, e):
		d = wx.MessageDialog(self, " A utility for calculating the real part of soft X-ray spectra written by Dr. Benjamin Watts at the Paul Scherrer Institute. If you make use of this utility, please consider citing the associated article:\n\nBenjamin Watts, ""Calculation of the Kramers-Kronig transform of X-ray spectra by a piecewise Laurent polynomial method,"" Opt. Express 22, 23628-23639 (2014)", "About KKcalc", wx.OK)
		# Create a message dialog box
		d.ShowModal() # Shows it
		d.Destroy() # finally destroy it when finished.

	def OnExit(self, e):
		self.Close(True)  # Close the frame.

	def OnOpen(self, e):
		"""Load data from a file."""
		success = False
		dlg = wx.FileDialog(self, "Choose a file", self.dirname, "", "*.*", wx.FD_OPEN)
		if dlg.ShowModal() == wx.ID_OK:
			success = True
			self.dirname, self.filename = os.path.split(dlg.GetPath())
		dlg.Destroy()
		if success:
			self.FileText.SetLabel("File: "+self.filename)
			self.raw_file = data.load_data(os.path.join(self.dirname, self.filename))
			self.combine_data()
			self.plot_data()

	def OnHelp(self, e):
		logger.info("Opening web browser for help files.")
		import webbrowser
		webbrowser.open(os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))),"README.rst"))
		#webbrowser.open("https://doi.org/10.1364/OE.22.023628")#README.rst")

	def OnArticle(self, e):
		logger.info("Opening web browser for article.")
		import webbrowser
		webbrowser.open("https://doi.org/10.1364/OE.22.023628")

	def OnSave(self, e):
		"""Write data to file."""
		convert_to = None
		if e.Id == 201:
			convert_to = "photoabsorption"
		elif e.Id == 202:
			convert_to = "refractive_index"
		logger.info("Save")
		fd = wx.FileDialog(self, style=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
		if fd.ShowModal()==wx.ID_OK:
			metadata = {"Density": float(self.DensityText.GetValue()), "Molecular Formula":self.StoichiometryText.GetValue(),"Formula Mass":data.calculate_FormulaMass(self.Stoichiometry)}
			data.export_data(fd.GetPath(), numpy.transpose(numpy.vstack((self.Full_E,self.KK_Real_Spectrum,data.coeffs_to_ASF(self.Full_E,self.Imaginary_Spectrum)))), header_info=metadata, convert_to=convert_to)

	def combine_data(self):
		"""Combine users near-edge data with extended spectrum data."""
		self.Full_E = None
		self.Imaginary_Spectrum = None
		if self.raw_file is not None:
			logger.info("Convert to scattering factors")
			self.NearEdgeData = data.convert_data(self.raw_file,self.DataTypeCombo.GetValue(),'ASF')
#			if self.InvertDataCheckBox.GetValue():
#				self.NearEdgeData[:,1] =  numpy.abs(self.NearEdgeData[:,1] - 2*numpy.mean(self.NearEdgeData[:,1]))
		logger.info("Combine Data")
		# Get splice points
		splice_eV = numpy.array([10.0, 30000.0])  # Henke limits
		try:
			splice_eV[0] = float(self.SpliceText1.GetValue())
		except ValueError:
			self.SpliceText1.SetValue("Start")
			if self.raw_file is not None:
				splice_eV[0] = self.NearEdgeData[0, 0]
		try:
			splice_eV[1] = float(self.SpliceText2.GetValue())
		except ValueError:
			self.SpliceText2.SetValue("End")
			if self.raw_file is not None:
				splice_eV[1] = self.NearEdgeData[-1, 0]
		# Merge the spectra
		if self.raw_file is not None and self.ASF_Data is None:
			self.Full_E, self.Imaginary_Spectrum, self.NearEdgeData, self.splice_ind  = data.merge_spectra(self.NearEdgeData, self.ASF_E, self.ASF_Data, merge_points=splice_eV, add_background=self.AddBackgroundCheckBox.GetValue(), plotting_extras=True)

		elif self.raw_file is None and self.ASF_Data is not None:
			self.Full_E = self.ASF_E
			self.Imaginary_Spectrum = self.ASF_Data

		elif self.raw_file is not None and self.ASF_Data is not None:
			
			self.Full_E, self.Imaginary_Spectrum, self.NearEdgeData, self.splice_ind  = data.merge_spectra(self.NearEdgeData, self.ASF_E, self.ASF_Data, merge_points=splice_eV, add_background=self.AddBackgroundCheckBox.GetValue(), fix_distortions=self.FixDistortionsCheckBox.GetValue(), plotting_extras=True)


	def plot_data(self):
		"""Plot data.
		
		Parameters:
		-----------
		self.Full_E : vector of floats 
			photon energies at which the real and imaginary scattering factor data will be plotted.
		self.Imaginary_Spectrum : Array of float 
			polynomial coefficients that can be evaluated to give the values of the imaginary scattering factors.
		self.KK_Real_Spectrum : vector of float 
			the values of the real scattering factors.

		Returns
		-------
		The GUI is updated, but nothing is returned.
		"""
		logger.info("plotting data")
		# List of things to plot
		plotlist = []
		# get initial guess at X limits
		X_min = 0
		X_max = 30000
		Y_max = 1
		Y_min = 0
		if self.NearEdgeData is not None:
			X_min = self.NearEdgeData[0, 0]
			X_max = self.NearEdgeData[-1, 0]
		if self.SpliceText1.GetValue() != "Start":
			X_min = float(self.SpliceText1.GetValue())
		if self.SpliceText2.GetValue() != "End":
			X_max = float(self.SpliceText2.GetValue())
		if self.Imaginary_Spectrum is not None:
			if self.Stoichiometry is not None:
				scale = sum([Z*count for Z, count in self.Stoichiometry])
			else:
				scale = 1.
			Im_energies, Im_values = data.coeffs_to_linear(self.Full_E, self.Imaginary_Spectrum, 0.001*scale)
			plotlist.append(plot.PolyLine(list(zip(Im_energies,Im_values)), colour='black', width=1))
			
			# get Y limits
			if self.splice_ind is None:
				Y_max = max(Im_values)
				Y_min = min(Im_values)
			else:
				Y_max = max(Im_values[self.splice_ind[0]:self.splice_ind[1]])
				Y_min = min(Im_values[self.splice_ind[0]:self.splice_ind[1]])
		if self.NearEdgeData is not None:
			Y_max = max(self.NearEdgeData[:,1])
			Y_min = min(self.NearEdgeData[:,1])
			plotlist.append(plot.PolyMarker(list(zip(self.NearEdgeData[:,0], self.NearEdgeData[:,1])), colour='blue', marker='plus', size=1))
			
		if self.splice_ind is not None:
			splice_values = data.coeffs_to_ASF(self.Full_E[self.splice_ind], self.Imaginary_Spectrum[[self.splice_ind[0],min(self.splice_ind[1],self.Imaginary_Spectrum.shape[0]-1)]])
			plotlist.append(plot.PolyMarker(list(zip(self.Full_E[self.splice_ind], splice_values)), colour='red', marker='cross', size=1))


		if self.raw_file is not None and self.Imaginary_Spectrum is None:
			logger.info("plot raw data only")
			plotlist.append(plot.PolyLine(self.NearEdgeData, colour='blue', width=1))  # User data
			if self.asf_bg is not None:
				plotlist.append(plot.PolyMarker(self.total_asf[self.asf_bg[0][0]:self.asf_bg[0][1], [0, 2]], colour='red', marker='cross', size=1))
				plotlist.append(plot.PolyLine(self.asf_bg[1], colour='red', width=1))

			# Real part
			#plotlist.append(plot.PolyLine(self.total_asf[:, [0, 1]], colour='black', width=1))
		if self.KK_Real_Spectrum is not None:
			if self.splice_ind is None:
				Y_max = max(self.KK_Real_Spectrum)
				Y_min = min(self.KK_Real_Spectrum)
			else:
				Y_max = max(Y_max, max(self.KK_Real_Spectrum[self.splice_ind[0]:self.splice_ind[1]]))
				Y_min = min(Y_min, min(self.KK_Real_Spectrum[self.splice_ind[0]:self.splice_ind[1]]))
			plotlist.append(plot.PolyLine(list(zip(self.Full_E, self.KK_Real_Spectrum)), colour='green', width=1))

		# Expand plotting limits for prettiness
		window_width = X_max-X_min
		X_max = X_max+window_width*0.1
		X_min = max(X_min-window_width*0.1, 0)
		window_Im_height = Y_max-Y_min
		window_Re_height = Y_max-Y_min
		Y_max = Y_max+window_Im_height*0.1
		Y_min = Y_min-window_Im_height*0.1
		Y_max = Y_max+window_Re_height*0.1
		Y_min = Y_min-window_Re_height*0.1
		# set up text, axis and draw
		#print plotlist
		#print X_min, X_max, Y_min, Y_max
		self.PlotAxes.Draw(plot.PlotGraphics(plotlist, '', 'Energy (eV)', 'Magnitude'), xAxis=(X_min, X_max), yAxis=(0, Y_max))
		#print "Plotlist =", len(plotlist)

	def Splice_Text_check(self, evt):
		self.combine_data()
		self.plot_data()

	def MergeAdd_check(self, evt):
		self.combine_data()
		self.plot_data()

	def Stoichiometry_Text_check(self, evt):
		if len(self.StoichiometryText.GetValue()) == 0:
			self.ChemicalFormula = None
			self.Stoichiometry = None
			self.Relativistic_Correction = None
			self.ASF_E = None
			self.ASF_Data = None
		else:
			self.ChemicalFormula = self.StoichiometryText.GetValue()
			self.Stoichiometry = data.ParseChemicalFormula(self.ChemicalFormula)
			self.Relativistic_Correction = kk.calc_relativistic_correction(self.Stoichiometry)
			self.ASF_E, self.ASF_Data = data.calculate_asf(self.Stoichiometry)
		self.combine_data()
		self.plot_data()



	def calculate(self, button):
		"""Calculate Button."""
		logger.debug("Calculate button")
		if self.Imaginary_Spectrum is not None:
			logger.info("Calculate Kramers-Kronig transform (PP)")
			self.KK_Real_Spectrum = kk.KK_PP(self.Full_E, self.Full_E, self.Imaginary_Spectrum, self.Relativistic_Correction)
			logger.info("Done!")
			self.plot_data()

def start_wx():
	app = wx.App()
	f = MyFrame()
	app.SetTopWindow(f)
	app.MainLoop()

if __name__ == '__main__':
	start_wx()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the Kramers-Kronig Calculator software package.
#
# Copyright (c) 2013 Benjamin Watts, Daniel J. Lauk
#
# The software is licensed under the terms of the zlib/libpng license.
# For details see LICENSE.txt

"""This module implements data formats and conversions."""

import logging
logger = logging.getLogger(__name__)
import json, os, warnings
import numpy, re


# Constants
CLASSICAL_ELECTRON_RADIUS = 2.81794029957951365441605230194258e-15  # meters
PLANCKS_CONSTANT = 4.1356673310e-15  # eV*seconds
SPEED_OF_LIGHT = 2.99792458e8  # meters per second
AVOGADRO_CONSTANT = 6.02214129e23  # no unit
LIST_OF_ELEMENTS = ['H','He','Li','Be','B','C','N','O','F','Ne','Na','Mg','Al','Si','P','S','Cl','Ar','K','Ca','Sc','Ti','V','Cr','Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In','Sn','Sb','Te','I','Xe','Cs','Ba','La','Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf','Ta','W','Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po','At','Rn','Fr','Ra','Ac','Th','Pa','U']

def load_Element_Database():
	"""Loads atomic scattering factor database from a json file.
	
	The database has been previously created by PackRawData.py
	
	Parameters
	----------
	None
	
	Returns
	-------
	The function returns a dictionary of elements, each consisting of a dictionary
	of data types

	"""
	logger.info("Loading element database")
	with open(os.path.join(os.path.dirname(__file__),'ASF.json'),'r') as f:
		Element_Database = json.load(f)
	for Z in range(1,93):
		Element_Database[str(Z)]['E'] = numpy.array(Element_Database[str(Z)]['E'])
		Element_Database[str(Z)]['Im'] = numpy.array(Element_Database[str(Z)]['Im'])
		Element_Database[str(Z)]['Re'] = numpy.array(Element_Database[str(Z)]['Re'])
	return Element_Database

ELEMENT_DATABASE = load_Element_Database() #spectral data, plus atomic masses

def coeffs_to_ASF(E, coeffs):
	"""Calculate Henke scattering factors from polynomial coefficients.

	Parameters
	----------
	E : float
		Energy in eV
	coeffs : array of floats
		PECS in cm^2/atom

	Returns
	-------
	The function returns the magnitude of the real or imaginary
	part of the atomic scattering factors at energy (or energies) `E`.

	"""
	if len(coeffs.shape) == 1:
		return coeffs[0]*E + coeffs[1] + coeffs[2]/E + coeffs[3]/(E**2) + coeffs[4]/(E**3)
	else:
		if len(E) == coeffs.shape[0]+1:
			coeffs = numpy.vstack((coeffs,coeffs[-1,:])) #Use last defined polynomial to calculate the ASF values at both second last and the last energies.
		return coeffs[:,0]*E + coeffs[:,1] + coeffs[:,2]/E + coeffs[:,3]/(E**2) + coeffs[:,4]/(E**3)

def convert_Beta_to_ASF(raw_data, density=None, formula_mass=None, stoichiometry=None, number_density=None, reverse=False):
	"""Convert Beta data (index of refraction) to atomic scattering
	factors (ASF).

	The Beta value is the imaginary part of the index of refraction.
	This represents the absorption.

	Parameters
	----------
	raw_data : two-dimensional `numpy.array` of `float`
		The array consists of two columns: Energy and magnitude.
	density : float
		material density in grams per millilitre.
	formula_mass : float
		atomic mass sum of the materials chemical formula (molecular mass).
	stoichiometry : a list of elemental symbol,number pairs
		description of the combination of elements composing the material.
	number_density : float
		material density in atoms per millilitre.
	reverse : boolean
		flag to indicate the reverse conversion.
	
	Note that the density in millilitres makes for a factor of a million in the equations.

	Returns
	-------
	The function returns a `numpy.array` of atomic scattering factors.
	They are made up of the energy and the magnitude of the imaginary
	part of the atomic scattering factors.

	"""
	if number_density is None:
		if density is None and formula_mass is None and stoichiometry is None:
			logger.info("Material number density required for conversions involving Beta.")
		if density is None:
			logger.info("Assuming a material mass density of 1 g/ml.")
			density = 1.0
		if formula_mass is None:
			if stoichiometry is not None:
				formula_mass = calculate_FormulaMass(stoichiometry)
			else:
				logger.info("Assuming a formula mass 100.")
				formula_mass = 100.
		number_density = density*AVOGADRO_CONSTANT/formula_mass
	if reverse: #ASF to Beta
		raw_Im = raw_data[:, :2].copy()
		raw_Im[:, 1] = 1000000.*raw_data[:, 1]*(number_density*CLASSICAL_ELECTRON_RADIUS*(PLANCKS_CONSTANT*SPEED_OF_LIGHT)**2)/(2*numpy.pi*raw_data[:, 0]**2)
	else: #Beta to ASF
		raw_Im = raw_data[:, :2].copy()
		raw_Im[:, 1] = 0.000001*raw_data[:, 1]*2*numpy.pi*raw_data[:, 0]**2/(number_density*CLASSICAL_ELECTRON_RADIUS*(PLANCKS_CONSTANT*SPEED_OF_LIGHT)**2)
	return raw_Im


def convert_BL_to_ASF(E, coeffs, Atomic_mass):
	"""Convert Biggs and Lighthill (BL) to atomic scattering factors (ASF).

	Biggs and Lighthill offers photoelectric cross-section (PECS) with the
	sum of AnE^-n for n=1-4  {E in keV and PECS in cm^2/g}.

	Henke scattering factors related by
	f2 = PECS*E/(2*r0*h*c)  {E in eV and PECS in cm^2/atom}.

	Parameters
	----------
	E : float
		Energy in eV
	coeffs : array of float
		Polynomial coefficients describing the function of the
		scattering factors
	Atomic_mass : float
		Atomic mass of the element (in units = g/mol)

	Returns
	-------
	The function returns the magnitude of the imaginary
	part of the atomic scattering factors at energy `E`.

	"""
	return (coeffs[0] + coeffs[1]/(E*0.001) + coeffs[2]/((E*0.001)**2) + coeffs[3]/((E*0.001)**3))*Atomic_mass/(2*AVOGADRO_CONSTANT*CLASSICAL_ELECTRON_RADIUS*PLANCKS_CONSTANT*SPEED_OF_LIGHT)*0.1


def convert_NEXAFS_to_ASF(raw_data, reverse=False):
	"""Convert NEXAFS photoabsorption data to atomic scattering factors (ASF).

	Parameters
	----------
	raw_data : two-dimensional `numpy.array` of `float`
		The array consists of two columns: Energy and magnitude.
	reverse : boolean
		flag to indicate the reverse conversion

	Returns
	-------
	The function returns a `numpy.array` of atomic scattering factors.
	They are made up of the energy and the magnitude of the imaginary
	part of the atomic scattering factors.

	"""
	if reverse:
		raw_Im = raw_data[:, :2].copy()
		raw_Im[:, 1] = (2*CLASSICAL_ELECTRON_RADIUS*PLANCKS_CONSTANT*SPEED_OF_LIGHT)*raw_data[:, 1]/raw_data[:, 0]
	else:
		raw_Im = raw_data[:, :2].copy()
		raw_Im[:, 1] = raw_data[:, 0]*raw_data[:, 1]/(2*CLASSICAL_ELECTRON_RADIUS*PLANCKS_CONSTANT*SPEED_OF_LIGHT)
	return raw_Im


def ConvertElementSymbol(SymbolString):
	"""Replace list of elemental symbols with the corresponding atomic numbers.

	Parameters
	----------
	SymbolString : String representing an elemental symbol

	Returns
	-------
	The function returns an integer atomic number corresponding to the input symbol.
	Zero is returned when the string is not recognised.
	"""
	try:
		AtomicNumber = LIST_OF_ELEMENTS.index(SymbolString)+1
	except ValueError:
		logger.warn("\""+SymbolString+"\" is not a known element!")
		AtomicNumber = 0
	return AtomicNumber

def ParseChemicalFormula(Formula,recursion_flag=False):
	"""Parse a chemical formula string to obtain a stoichiometry.

	Parameters
	----------
	Formula : string consisting of element symbols, numbers and parentheses

	Returns
	-------
	The function returns a list of elemental symbol,number pairs
	
	TODO: return understood formula (i.e. input strign with errors removed)
	"""
	if not recursion_flag:
		logger.info("Parsing '"+Formula+"' as a chemical formula")
	Stoichiometry = []
	m=re.search('((?P<Element>[A-Z][a-z]?)|\((?P<Paren>.*)\))(?P<Number>\d*(\.\d+)?)(?P<Remainder>.*)',Formula)
	if len(m.group('Number')) is not 0:
		Number = float(m.group('Number'))
	else:
		Number = 1.0
	if m.group('Element') is not None:
		Z = ConvertElementSymbol(m.group('Element'))
		if Z is not 0:
			Stoichiometry.append([Z,Number])
	elif len(m.group('Paren')) > 0:
		Stoichiometry +=[[x[0],x[1]*Number] for x in ParseChemicalFormula(m.group('Paren'),recursion_flag=True)]
	if len(m.group('Remainder')) is not 0:
		Stoichiometry += ParseChemicalFormula(m.group('Remainder'),recursion_flag=True)
	return Stoichiometry
	
def calculate_FormulaMass(Stoichiometry):
	"""Sum atomic masses
	Parameters
	-------
	Stoichiometry : a list of elemental symbol,number pairs

	Returns
	-------
	The function returns the fomula mass as a float
	"""
	FormulaMass = 0
	for element, number in Stoichiometry:
		FormulaMass += number*ELEMENT_DATABASE[str(element)]['mass']
	return FormulaMass
	
def load_data(filename, load_options=None):
	"""Read a standard ASCII file and return a list of lists of floats.
	
	Parameters
	----------
	filename : string path to data file
	load_options : dictionary of optional settings

	Returns
	-------
	The function returns a numpy array with two columns: Photon energy and Imaginary scattering factor values
	"""
	logger.info("Load data from file")
	data = []
	if os.path.isfile(filename):
		for line in open(filename):
			try:
				data.append([float(f) for f in line.split()])
			except ValueError:
				pass
		data = numpy.array(data)
	else:
		logger.error(filename+" is not a valid file name.")
	if len(data)==0:
		logger.error("no data found in "+filename)
		return None
	else:
		if isinstance(load_options,dict) and 'E_column' in load_options:
			E_column = int(load_options['E_column'])
		else:
			E_column = 0
		if isinstance(load_options,dict) and 'data_column' in load_options:
			data_column = int(load_options['data_column'])
		else:
			data_column = data.shape[1]-1
		return data[:,[E_column,data_column]]
	
def export_data(filename, data, header_info=None,convert_to=None):
	"""Write spectral data into an ASCII file.
	
	Parameters
	----------
	filename : string path to data file
	data : array of floats, arranged in columns, with energy in the first.
	save_options : dictionary of optional settings

	Returns
	-------
	None
	"""
	logger.info("Export data to file")
	if data is not None:
		column_headings = '# E(eV)\tf1\tf2\n'
		data_type = "Scattering factors"
		if convert_to == 'photoabsorption':
			data = convert_data(data[:,[0,2]],'ASF',convert_to, Density=header_info['Density'])
			column_headings = '# E(eV)\tphotoabsorption\n'
			data_type = "Photoabsorption spectrum"
		elif convert_to == 'refractive_index':
			data[:,1] = convert_data(data[:,[0,1]],'ASF',convert_to, Density=header_info['Density'], Formula_Mass=header_info["Formula Mass"])[:,1]
			data[:,2] = convert_data(data[:,[0,2]],'ASF',convert_to, Density=header_info['Density'], Formula_Mass=header_info["Formula Mass"])[:,1]
			column_headings = '# E(eV)\tDelta\tBeta\n'
			data_type = "Refractive index"
		with open(filename, "w") as outfile:
			if header_info is not None:
				for key in header_info:
					outfile.write('# '+key+' = '+str(header_info[key])+'\n')
			outfile.write(column_headings)
			numpy.savetxt(outfile,data,fmt="%7g",delimiter="\t")
		logger.info(data_type+" for "+header_info["Molecular Formula"]+" saved to "+filename)
	else:
		logger.info("Nothing to save.")









def convert_data(Data, FromType, ToType, Density=None, Formula_Mass=None):
	"""Switchyard function for converting between data types.
	
	Parameters
	----------
	Data : a numpy array with two columns: Photon energy and absorption data values
	FromType : string describing input data type
	ToType : string describing desired output data type

	Returns
	-------
	The function returns a numpy array with two columns: Photon energy and absorption data values
	"""
	logger.info("Convert data from "+FromType+" to "+ToType+".")
	if FromType.lower() in ['photoabsorption', 'nexafs', 'xanes']:
		converted_data = convert_NEXAFS_to_ASF(Data)
	elif FromType.lower() in ['beta']:
		converted_data = convert_Beta_to_ASF(Data, density=Density)
	else: #if already ASF
		converted_data = Data.copy()
	#data should now be in terms of ASF
	if ToType.lower() in ['photoabsorption', 'nexafs', 'xanes']:
		converted_data = convert_NEXAFS_to_ASF(converted_data, reverse=True)
	elif ToType.lower() in ['refractive_index','beta']:
		converted_data = convert_Beta_to_ASF(converted_data, density=Density, formula_mass=Formula_Mass, reverse=True)
	return converted_data
	
	
def calculate_asf(Stoichiometry):
	"""Sum scattering factor data for a given chemical stoichiometry.
	
	Parameters
	----------
	Stoichiometry : a list of elemental symbol,number pairs
	
	Returns
	-------
	total_E: 1D numpy array listing the starting photon energies of the segments that the spectrum is broken up into.
	total_Im_coeffs: nx5 numpy array in which each row lists the polynomial coefficients describing the shape of the spectrum in that segment.
	"""
	logger.info("Calculate material scattering factor data from the given stoichiometry")
	if len(Stoichiometry) is 0:
		logger.error("No elements described by input.")
		return None
	else:
		# get unique energy points
		total_E = numpy.array([])
		for element,n in Stoichiometry:
			total_E = numpy.concatenate((total_E, ELEMENT_DATABASE[str(element)]['E']))
		total_E = numpy.unique(total_E)
		# add weighted asf data sets for KK calculation
		total_Im_coeffs = numpy.zeros((len(total_E)-1, 5))
		counters = numpy.zeros((len(Stoichiometry)),dtype=numpy.int)
		for i,E in enumerate(total_E[1:]):
			sum_Im_coeffs = 0
			for j in range(len(counters)):
				sum_Im_coeffs += Stoichiometry[j][1]*ELEMENT_DATABASE[str(Stoichiometry[j][0])]['Im'][counters[j],:]
				counters[j] += ELEMENT_DATABASE[str(Stoichiometry[j][0])]['E'][counters[j]+1] == E
			total_Im_coeffs[i,:] = sum_Im_coeffs
		return total_E, total_Im_coeffs

def merge_spectra(NearEdge_Data, ASF_E, ASF_Data, merge_points=None, add_background=False, fix_distortions=False, plotting_extras=False):
	"""Normalise the user-provided, near-edge data to the scattering factor data and combine them.
	
	Parameters
	----------
	NearEdge_Data : a numpy array with two columns: Photon energy and absorption data values.
	ASF_E : 1D numpy array listing the starting photon energies for each spectrum segment.
	ASF_Data: nx5 numpy array in which each row lists the polynomial coefficients describing the shape of the spectrum in that segment.
	merge_points : a pair of photon energies indicating the data range of the NearEdge_Data to be used.
	add_background : boolean flag for adding a background function to the provided near-edge data.
	fix_distortions : boolean flag for removing erroneous slope from the provided near-edge data.
	plotting_extras : boolean flag for providing plottable feedback data that isn't required for actual calculations.

	Returns
	-------
	Full_E : 1D numpy array listing the updated starting photon energies for each spectrum segment.
	Full_Coeffs : nx5 numpy array in which each row lists the polynomial coefficients describing the shape of the spectrum in that segment.
	plot_scaled_NearEdge_Data (optional) : an updated verion of NearEdge_Data that has been scaled to match the scattering factor data.
	splice_points (optional) : a list of two pairs of photon energy-magnitude values to represent the merge_points in a plot.
	"""
	logger.info("Merge near-edge data with wide-range scattering factor data")
	if merge_points is None:
		merge_points = NearEdge_Data[[0,-1],0]
	NearEdge_merge_ind = [numpy.where(NearEdge_Data[:,0] > merge_points[0])[0][0], numpy.where(NearEdge_Data[:,0] < merge_points[1])[0][-1]]
	NearEdge_merge_values = numpy.interp(merge_points, NearEdge_Data[:,0], NearEdge_Data[:,1])
	if ASF_Data is not None:
		asf_merge_ind = [numpy.where(ASF_E > merge_points[0])[0][0]-1, numpy.where(ASF_E > merge_points[1])[0][0]-1]
		asf_merge_values = [coeffs_to_ASF(merge_points[0], ASF_Data[asf_merge_ind[0],:]), coeffs_to_ASF(merge_points[1], ASF_Data[asf_merge_ind[1],:])]
		if add_background:
			logger.info("Add background")
			logger.error("Not implemented!")
			#get pre-edge region
			#extrapolate background
			scale = (asf_merge_values[1]-asf_merge_values[0])/(NearEdge_merge_values[1]-NearEdge_merge_values[0])
			scaled_NearEdge_Data = numpy.vstack((NearEdge_Data[:,0],((NearEdge_Data[:, 1]-NearEdge_merge_values[0])*scale)+asf_merge_values[0])).T
		else:# don't add background
			scale = (asf_merge_values[1]-asf_merge_values[0])/(NearEdge_merge_values[1]-NearEdge_merge_values[0])
			scaled_NearEdge_Data = numpy.vstack((NearEdge_Data[:,0],((NearEdge_Data[:, 1]-NearEdge_merge_values[0])*scale)+asf_merge_values[0])).T
		try:
			import scipy.optimize
			SCIPY_FLAG = True
		except ImportError:
			SCIPY_FLAG = False
			logger.info('Failed to import the scipy.optimize module - disabling the \'fix distortions\' option.')
		if SCIPY_FLAG and fix_distortions:
			logger.info("Fix distortions")
			import scipy.optimize
			ASF_fitY = 0.0*NearEdge_Data[:, 0]
			for i, E in enumerate(NearEdge_Data[:,0]):
				ASF_fitY[i] = coeffs_to_ASF(E, ASF_Data[numpy.where(ASF_E > E)[0][0]-1])
			fitfunc = lambda p, x, y, asf_mv, asf: ((y-p*x)-(y[0]-p*x[0]))/((y[-1]-p*x[-1])-(y[0]-p*x[0]))*(asf_mv[1]-asf_mv[0])+asf_mv[0] - asf
			p0 = -(NearEdge_merge_values[1]-NearEdge_merge_values[0])/((asf_merge_values[1]-asf_merge_values[0])*NearEdge_Data[0,0])
			logger.debug("Fix distortions - start fit with p0 ="+str(p0))
			p1, success = scipy.optimize.leastsq(fitfunc, p0, args=(NearEdge_Data[:, 0], NearEdge_Data[:, 1], asf_merge_values, ASF_fitY))
			logger.debug("Fix distortions - complete fit with p1 ="+str(p1[0]))
			NearEdge_fitY = asf_merge_values[0]+((NearEdge_Data[:,1]-p1[0]*NearEdge_Data[:,0])-(NearEdge_Data[0,1]-p1[0]*NearEdge_Data[0,0]))*(asf_merge_values[1]-asf_merge_values[0])/((NearEdge_Data[-1,1]-p1[0]*NearEdge_Data[-1,0])-(NearEdge_Data[0,1]-p1[0]*NearEdge_Data[0,0]))
			scaled_NearEdge_Data = numpy.vstack((NearEdge_Data[:,0],NearEdge_fitY)).T
	else:
		scaled_NearEdge_Data = NearEdge_Data.copy()
		asf_merge_values = NearEdge_merge_values.copy()
	plot_scaled_NearEdge_Data = scaled_NearEdge_Data.copy()
	scaled_NearEdge_Data = numpy.vstack(((merge_points[0],asf_merge_values[0]),scaled_NearEdge_Data[NearEdge_merge_ind[0]:NearEdge_merge_ind[1]+1,:],(merge_points[1],asf_merge_values[1])))

	#Now convert point values to coefficients
	NearEdge_Coeffs = numpy.zeros((len(scaled_NearEdge_Data),5))
	M = (scaled_NearEdge_Data[1:,1]-scaled_NearEdge_Data[:-1,1])/(scaled_NearEdge_Data[1:,0]-scaled_NearEdge_Data[:-1,0])
	NearEdge_Coeffs[:-1,0] = M
	NearEdge_Coeffs[:-1,1] = scaled_NearEdge_Data[:-1,1]-M*scaled_NearEdge_Data[:-1,0]
	if ASF_Data is None:
		NearEdge_Coeffs = NearEdge_Coeffs[0:-1,:]
		#NearEdge_Coeffs[-1,:] = NearEdge_Coeffs[-2,:]
		#Paste data sections together
		Full_E = scaled_NearEdge_Data[:,0]
		Full_Coeffs = NearEdge_Coeffs
		splice_points = [0,len(Full_E)-1]
	else:
		NearEdge_Coeffs[-1,:] = ASF_Data[asf_merge_ind[1],:]
		#Paste data sections together
		Full_E = numpy.hstack((ASF_E[0:asf_merge_ind[0]+1],scaled_NearEdge_Data[:,0],ASF_E[asf_merge_ind[1]+1:]))
		Full_Coeffs = numpy.vstack((ASF_Data[0:asf_merge_ind[0]+1,:],NearEdge_Coeffs,ASF_Data[asf_merge_ind[1]+1:,:]))
		splice_points = [asf_merge_ind[0]+1, asf_merge_ind[0]+len(scaled_NearEdge_Data[:,0])]
	if plotting_extras:
		return Full_E, Full_Coeffs, plot_scaled_NearEdge_Data, splice_points
	else:
		return Full_E, Full_Coeffs
	
	
def coeffs_to_linear(E, coeffs, threshold):
	"""Convert a curved data set (described by polynomial coefficients) to an approximate data set composed of linear segments.
	   This should be useful for plotting
	
	Parameters
	----------
	E : 1D numpy array listing the starting photon energies for each spectrum segment.
	coeffs: nx5 numpy array in which each row lists the polynomial coefficients describing the shape of the spectrum in that segment.
	threshold: scalar value indicating the maximum amount of error to be tolerated.
	
	Returns
	-------
	linear_E: 1D numpy array listing the starting photon energies of the segments that the spectrum is broken up into.
	linear_Vals: 1D numpy array listing the  intensity values corresponding to the energies listed in linear_E.
	"""
	logger.info("Linearise data for plotting.")
	curvature = 2*coeffs[:,2]/(E[0:-1]**3) + 6*coeffs[:,3]/(E[0:-1]**4) + 12*coeffs[:,4]/(E[0:-1]**5)
	linear_E = numpy.array([])
	linear_Vals = numpy.array([])
	for i in range(len(E)-1):
		linear_E = numpy.append(linear_E, E[i])
		linear_Vals = numpy.append(linear_Vals, coeffs_to_ASF(E[i],coeffs[i]))
		if not numpy.all(coeffs[i,2::] == [0,0,0]):
			new_points = split_segment_recursively(E[i], E[i+1], coeffs[i], threshold)
			#print new_points
			if new_points is not None:
				for p in new_points:
					linear_E = numpy.append(linear_E, p[0])
					linear_Vals = numpy.append(linear_Vals, p[1])
	linear_E = numpy.append(linear_E, E[-1])
	linear_Vals = numpy.append(linear_Vals, coeffs_to_ASF(E[-1],coeffs[-1]))
	return linear_E, linear_Vals

def split_segment_recursively(E1, E2, coeffs, threshold):
	"""Recursively use split_segment() to divide a spectrum segment until all points along the piecewise linear representation
	is within 'threshold' of the curve defined by 'coeffs'.
	
	Parameters
	----------
	E1 : Energy value at the start of the segment.
	E2 : Energy value at the end of the segment.
	coeffs: 1x5 numpy array listing the polynomial coefficients describing the shape of the spectrum in that segment.
	threshold: scalar value indicating the maximum amount of error to be tolerated.
	
	Returns
	-------
	The function returns a list of energy, magnitude tuples representing an approximation of the curve described by 'coeffs'.
	"""
	new_point = split_segment(E1, E2, coeffs)
	if new_point is not None and new_point[2] > threshold:
		#print new_point
		point_list = []
		new_point_2 = split_segment_recursively(E1, new_point[0], coeffs, threshold)#recursively check first half
		if new_point_2 is not None:
			point_list = point_list + new_point_2
		point_list.append(new_point[0:2])
		new_point_3 = split_segment_recursively(new_point[0], E2, coeffs, threshold)#recursively check second half
		if new_point_3 is not None:
			point_list = point_list + new_point_3
		return point_list
	else:
		return None

def split_segment(E1, E2, coeffs):
	"""Calculate which point in a spectrum segment shows the greatest error when represented by a straight line.
	
	Parameters
	----------
	E1 : Energy value at the start of the segment.
	E2 : Energy value at the end of the segment.
	coeffs: 1x5 numpy array listing the polynomial coefficients describing the shape of the spectrum in that segment.
	
	Returns
	-------
	The function returns the energy, magnitude and error value corresponding to the point at which the greatest error occurs.
	"""
	vals = [coeffs_to_ASF(E1, coeffs), coeffs_to_ASF(E2, coeffs)]
	m = (vals[1] - vals[0])/(E2 - E1)
	a = m-coeffs[0]
	b = 0.0
	c = coeffs[2]
	d = 2*coeffs[3]
	e = 3*coeffs[4]
	
	# see http://en.wikipedia.org/wiki/Quartic_function#General_formula_for_roots
	D0 = c**2-3*b*d+12*a*e
	D1 = 2*c**3-9*b*c*d+27*b**2*e+27*a*d**2-72*a*c*e
	p = (8*a*c-3*b**2)/(8.0*a**2)
	q = (b**3-4*a*b*c+8*a**2*d)/(8.0*a**3)
	Q_cubed = (D1+(D1**2-4*D0**3)**0.5)*0.5
	Q = numpy.sign(Q_cubed)*abs(Q_cubed)**(1/3.0)
	S = 0.5*(-3*p/2.0+1/(3.0*a)*(Q+D0/Q))**0.5
	with warnings.catch_warnings():
		warnings.simplefilter("ignore")
		#This line will generate warnings when failing to calculate complex roots properly, but we only care about real roots anyway. 
		roots = numpy.array([-b/(4.0*a)-S+0.5*(-4*S**2-2*p+q/S)**0.5, -b/(4.0*a)-S-0.5*(-4*S**2-2*p+q/S)**0.5, -b/(4.0*a)+S+0.5*(-4*S**2-2*p-q/S)**0.5, -b/(4.0*a)+S-0.5*(-4*S**2-2*p-q/S)**0.5])
	good_roots = []
	for r in roots:
		if numpy.isfinite(r) & (E1<r<E2):
			good_roots.append(r)
	if len(good_roots) == 0:
		return None
	elif len(good_roots) == 1:
		new_E = good_roots[0]
		distance = abs(vals[0]+(new_E-E1)*m - coeffs_to_ASF(new_E, coeffs))
	else:
		distance = abs(vals[0]+(good_roots-E1)*m - coeffs_to_ASF(good_roots, coeffs))
		best_root_ind = numpy.argmin(distance)
		new_E = good_roots[best_root_ind]
		distance = distance[best_root_ind]
	return new_E, coeffs_to_ASF(new_E, coeffs), distance


	



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the Kramers-Kronig Calculator software package.
#
# Copyright (c) 2013 Benjamin Watts, Daniel J. Lauk
#
# The software is licensed under the terms of the zlib/libpng license.
# For details see LICENSE.txt

"""This module implements the Kramers-Kronig transformation."""

import logging, sys
logger = logging.getLogger(__name__)
if __name__ == '__main__':
	logging.basicConfig(level=logging.DEBUG)
	logging.StreamHandler(stream=sys.stdout)

import math
import numpy
import os
import data


def calc_relativistic_correction(stoichiometry):
	"""Calculate the relativistic correction to the Kramers-Kronig transform.

	Parameters:
	-----------
	stoichiometry : array of integer/float pairs
		Each pair in the list consists of an atomic number and the relative proportion of that element.

	Returns
	-------
	This function returns a ``float`` holding the relativistic
	corection to the Kramers-Kronig transform.

	"""
	correction = 0
	for z, n in stoichiometry:
		correction += (z - (z/82.5)**2.37) * n
	return correction


def KK_General_PP(Eval_Energy, Energy, imaginary_spectrum, orders, relativistic_correction):
	"""Calculate Kramers-Kronig transform with "Piecewise Polynomial"
	algorithm plus the Biggs and Lighthill extended data.

	Parameters
	----------
	Eval_Energy : numpy vector of `float`
		Set of photon energies describing points at which to evaluate the real spectrum
	Energy : numpy vector of `float`
		Set of photon energies describing intervals for which each row of `imaginary_spectrum` is valid
	imaginary_spectrum : two-dimensional `numpy.array` of `float`
		The array consists of columns of polynomial coefficients belonging to the power terms indicated by 'order'
	orders : numpy vector of integers
		The vector represents the polynomial indices corresponding to the columns of imaginary_spectrum
	relativistic_correction : float
		The relativistic correction to the Kramers-Kronig transform.
		You can calculate the value using the `calc_relativistic_correction` function.

	Returns
	-------
	This function returns the real part of the scattering factors evaluated at photon energies specified by Eval_Energy.

	"""
	logger = logging.getLogger(__name__)
	logger.info("Calculate Kramers-Kronig transform using general piecewise-polynomial algorithm")
	# Need to build x-E-n arrays
	X = numpy.tile(Energy[:,numpy.newaxis,numpy.newaxis],(1,len(Eval_Energy),len(orders)))
	E = numpy.tile(Eval_Energy[numpy.newaxis,:,numpy.newaxis],(len(Energy)-1,1,len(orders)))
	C = numpy.tile(imaginary_spectrum[:,numpy.newaxis,:],(1,len(Eval_Energy),1))
	N = numpy.tile(orders[numpy.newaxis,numpy.newaxis,:],(len(Energy)-1,len(Eval_Energy),1))
	poles = numpy.equal(X,numpy.tile(Eval_Energy[numpy.newaxis,:,numpy.newaxis],(len(Energy),1,len(orders))))
	
	# all N, ln(x+E) and ln(x-E) terms and poles
	Integral = numpy.sum(-C*(-E)**N*numpy.log(numpy.absolute((X[1:,:,:]+E)/(X[:-1,:,:]+E)))-C*E**N*(1-poles[1:,:,:])*numpy.log(numpy.absolute((X[1:,:,:]-E+poles[1:,:,:])/((1-poles[:-1,:,:])*X[:-1,:,:]+poles[:-1,:,:]*X[[0]+list(range(len(Energy)-2)),:,:]-E))),axis=(0,2))
	
	if numpy.any(orders<=-2): # N<=-2, ln(x) terms
		i = [slice(None,None,None),slice(None,None,None),orders<=-2]
		Integral += numpy.sum(C[i]*((-E[i])**N[i]+E[i]**N[i])*numpy.log(numpy.absolute((X[1:,:,orders<=-2])/(X[:-1,:,orders<=-2]))),axis=(0,2))
	
	if numpy.any(orders>=0): # N>=0,  x^k terms
		for ni in numpy.where(orders>=0)[0]:
			i = [slice(None,None,None),slice(None,None,None),ni]
			n = orders[ni]
			for k in range(n,0,-2):
				Integral += numpy.sum(C[i]/float(-k)*2*E[i]**(n-k)*(X[1:,:,ni]**k-X[:-1,:,ni]**k),axis=0)
	
	if numpy.any(orders <=-3): # N<=-3, x^k terms
		for ni in numpy.where(orders<=-3)[0]:
			i = [slice(None,None,None),slice(None,None,None),ni]
			n = orders[ni]
			for k in range(n+2,0,2):
				Integral += numpy.sum(C[i]/float(k)*((-1)**(n-k)+1)*E[i]**(n-k)*(X[1:,:,ni]**k-X[:-1,:,ni]**k),axis=0)
	
	logger.debug("Done!")
	return Integral / math.pi + relativistic_correction

def KK_PP(Eval_Energy, Energy, imaginary_spectrum, relativistic_correction):
	"""Calculate Kramers-Kronig transform with "Piecewise Polynomial"
	algorithm plus the Biggs and Lighthill extended data.

	Parameters
	----------
	Eval_Energy : numpy vector of `float`
		Set of photon energies describing points at which to evaluate the real spectrum
	Energy : numpy vector of `float`
		Set of photon energies describing intervals for which each row of `imaginary_spectrum` is valid
	imaginary_spectrum : two-dimensional `numpy.array` of `float`
		The array consists of five columns of polynomial coefficients: A_1, A_0, A_-1, A_-2, A_-3
	relativistic_correction : float
		The relativistic correction to the Kramers-Kronig transform.
		You can calculate the value using the `calc_relativistic_correction` function.

	Returns
	-------
	This function returns the real part of the scattering factors evaluated at photon energies specified by Eval_Energy.

	"""
	logger = logging.getLogger(__name__)
	logger.info("Calculate Kramers-Kronig transform using (n from 1 to -3) piecewise-polynomial algorithm")
	X1 = Energy[0:-1]
	X2 = Energy[1:]
	E = numpy.tile(Eval_Energy, (len(Energy)-1, 1)).T
	Full_coeffs = imaginary_spectrum.T
	Symb_1 = (( Full_coeffs[0, :]*E+Full_coeffs[1, :])*(X2-X1)+0.5*Full_coeffs[0, :]*(X2**2-X1**2)-(Full_coeffs[3, :]/E+Full_coeffs[4, :]*E**-2)*numpy.log(numpy.absolute(X2/X1))+Full_coeffs[4, :]/E*(X2**-1-X1**-1))
	Symb_2 = ((-Full_coeffs[0, :]*E+Full_coeffs[1, :])*(X2-X1)+0.5*Full_coeffs[0, :]*(X2**2-X1**2)+(Full_coeffs[3, :]/E-Full_coeffs[4, :]*E**-2)*numpy.log(numpy.absolute(X2/X1))-Full_coeffs[4, :]/E*(X2**-1-X1**-1))+(Full_coeffs[0, :]*E**2-Full_coeffs[1, :]*E+Full_coeffs[2, :]-Full_coeffs[3, :]*E**-1+Full_coeffs[4, :]*E**-2)*numpy.log(numpy.absolute((X2+E)/(X1+E)))
	Symb_3 = (1-1*((X2==E)|(X1==E)))*(Full_coeffs[0, :]*E**2+Full_coeffs[1, :]*E+Full_coeffs[2, :]+Full_coeffs[3, :]*E**-1+Full_coeffs[4, :]*E**-2)*numpy.log(numpy.absolute((X2-E+1*(X2==E))/(X1-E+1*(X1==E))))
	Symb_B = numpy.sum(Symb_2 - Symb_1 - Symb_3, axis=1)  # Sum areas for approximate integral
	# Patch singularities
	hits = Energy[1:-1]==E[:,0:-1]
	E_hits = numpy.append(numpy.insert(numpy.any(hits, axis=0),[0,0],False),[False,False])
	Eval_hits = numpy.any(hits, axis=1)
	X1 = Energy[E_hits[2:]]
	XE = Energy[E_hits[1:-1]]
	X2 = Energy[E_hits[:-2]]
	C1 = Full_coeffs[:, E_hits[2:-1]]
	C2 = Full_coeffs[:, E_hits[1:-2]]
	Symb_singularities = numpy.zeros(len(Eval_Energy))
	Symb_singularities[Eval_hits] = (C2[0, :]*XE**2+C2[1, :]*XE+C2[2, :]+C2[3, :]*XE**-1+C2[4, :]*XE**-2)*numpy.log(numpy.absolute((X2-XE)/(X1-XE)))
	# Finish things off
	KK_Re = (Symb_B-Symb_singularities) / (math.pi*Eval_Energy) + relativistic_correction
	logger.debug("Done!")
	return KK_Re

def improve_accuracy(Full_E, Real_Spectrum, Imaginary_Spectrum, relativistic_correction, tolerance, recursion=50):
	"""Calculate extra data points so that a linear interpolation is more accurate.
	
	Parameters
	----------
	Full_E : numpy vector of `float`
		Set of photon energies describing intervals for which each row of `imaginary_spectrum` is valid
	Real_Spectrum : numpy vector of `float`
		The real part of the spectrum corresponding to magnitudes at photon energies in Full_E
	Imaginary_Spectrum : two-dimensional `numpy.array` of `float`
		The array consists of five columns of polynomial coefficients: A_1, A_0, A_-1, A_-2, A_-3
	relativistic_correction : float
		The relativistic correction to the Kramers-Kronig transform.
		(You can calculate the value using the `calc_relativistic_correction` function.)
	tolerance : float
		Level of error in linear extrapolation of data values to be allowed.
	recursion : integer
		Number of times an energy interval can be halved before giving up.

	Returns
	-------
	This function returns a numpy array with three columns respectively representing photon energy, the real spectrum and the imaginary spectrum.
	"""
	logger.debug("Improve data accuracy")
	new_points = numpy.cumsum(numpy.ones((len(Full_E)-2,1),dtype=numpy.int8))+1
	Im_values = data.coeffs_to_ASF(Full_E, numpy.vstack((Imaginary_Spectrum,Imaginary_Spectrum[-1])))
	#plot_Im_values = Im_values
	Re_values = Real_Spectrum
	E_values = Full_E
	temp_Im_spectrum = Imaginary_Spectrum[1:]
	count = 0
	improved = 1
	total_improved_points = 0
	while count<recursion and numpy.sum(improved)>0:
		#get E_midpoints
		midpoints = (E_values[new_points-1]+E_values[new_points])/2.
		#evaluate at new points
		Im_midpoints = data.coeffs_to_ASF(midpoints, temp_Im_spectrum)
		Re_midpoints = KK_PP(midpoints, Full_E, Imaginary_Spectrum, relativistic_correction)
		#evaluate error levels
		Im_error = abs((Im_values[new_points-1]+Im_values[new_points])/2. - Im_midpoints)
		Re_error = abs((Re_values[new_points-1]+Re_values[new_points])/2. - Re_midpoints)
		improved = (Im_error>tolerance) | (Re_error>tolerance)
		logger.debug(str(numpy.sum(improved))+" points (out of "+str(len(improved))+") can be improved in pass number "+str(count+1)+".")
		total_improved_points += numpy.sum(improved)
		#insert new points and values
		Im_values = numpy.insert(Im_values,new_points[improved],Im_midpoints[improved])
		Re_values = numpy.insert(Re_values,new_points[improved],Re_midpoints[improved])
		E_values = numpy.insert(E_values,new_points[improved],midpoints[improved])
		#prepare for next loop
		temp_Im_spectrum =numpy.repeat(temp_Im_spectrum[improved],2,axis=0)
		new_points = numpy.where(numpy.insert(numpy.zeros(Im_values.shape, dtype=numpy.bool),new_points[improved],True))[0]
		new_points = numpy.vstack((new_points, new_points+1)).T.flatten()
		count += 1
	
	#import matplotlib
	#matplotlib.use('WXAgg')
	#import pylab
	#pylab.figure()
	#pylab.plot(Full_E,plot_Im_values,'ok')
	#pylab.plot(Full_E,Real_Spectrum,'og')
	#pylab.plot(midpoints,Im_midpoints,'+b')
	#pylab.plot(midpoints,Re_midpoints,'+r')
	#pylab.plot(E_values,Im_values,'b-')
	#pylab.plot(E_values,Re_values,'r-')
	#pylab.plot(midpoints,Im_error,'b-')
	#pylab.plot(midpoints,Re_error,'r-')
	#pylab.xscale('log')
	#pylab.show()
	logger.info("Improved data accuracy by inserting "+str(total_improved_points)+" extra points.")
	return numpy.vstack((E_values,Re_values,Im_values)).T
	
def kk_calculate_real(NearEdgeDataFile, ChemicalFormula, load_options=None, input_data_type=None, merge_points=None, add_background=False, fix_distortions=False, curve_tolerance=None, curve_recursion=50):
	"""Do all data loading and processing and then calculate the kramers-Kronig transform.
	Parameters
	----------
	NearEdgeDataFile : string
		Path to file containg near-edge data
	ChemicalFormula : string
		A standard chemical formula string consisting of element symbols, numbers and parentheses.
	merge_points : list or tuple pair of `float` values, or None
		The photon energy values (low, high) at which the near-edge and scattering factor data values
		are set equal so as to ensure continuity of the merged data set.

	Returns
	-------
	This function returns a numpy array with columns consisting of the photon energy, the real and the imaginary parts of the scattering factors.
	"""
	Stoichiometry = data.ParseChemicalFormula(ChemicalFormula)
	Relativistic_Correction = calc_relativistic_correction(Stoichiometry)
	Full_E, Imaginary_Spectrum = data.calculate_asf(Stoichiometry)
	if NearEdgeDataFile is not None:
		NearEdge_Data = data.convert_data(data.load_data(NearEdgeDataFile, load_options),FromType=input_data_type,ToType='asf')
		Full_E, Imaginary_Spectrum = data.merge_spectra(NearEdge_Data, Full_E, Imaginary_Spectrum, merge_points=merge_points, add_background=add_background, fix_distortions=fix_distortions)
	Real_Spectrum = KK_PP(Full_E, Full_E, Imaginary_Spectrum, Relativistic_Correction)
	if curve_tolerance is not None:
		output_data = improve_accuracy(Full_E,Real_Spectrum,Imaginary_Spectrum, Relativistic_Correction, curve_tolerance, curve_recursion)
	else:
		Imaginary_Spectrum_Values = data.coeffs_to_ASF(Full_E, numpy.vstack((Imaginary_Spectrum,Imaginary_Spectrum[-1])))
		output_data = numpy.vstack((Full_E,Real_Spectrum,Imaginary_Spectrum_Values)).T
	return output_data

if __name__ == '__main__':
	#use argparse here to get command line arguments
	#process arguments and pass to a pythonic function
	
	#I will abuse this section of code for initial testing
	#Output = kk_calculate_real('../../data/Xy_norm_bgsub.txt', 'C10SH14', input_data_type='NEXAFS')
	Output = kk_calculate_real('../../data/LaAlO3/LaAlO3_Exp.csv', 'LaAlO3', input_data_type='NEXAFS', fix_distortions=True, curve_tolerance=0.05)
	#Output = kk_calculate_real('../../data/GaAs/As.xmu.csv', 'GaAs', input_data_type='NEXAFS', fix_distortions=True, curve_tolerance=0.05)
	
	Stoichiometry = data.ParseChemicalFormula('LaAlO3')
	#Stoichiometry = data.ParseChemicalFormula('GaAs')
	Relativistic_Correction = calc_relativistic_correction(Stoichiometry)
	ASF_E, ASF_Data = data.calculate_asf(Stoichiometry)
	ASF_Data3 = data.coeffs_to_linear(ASF_E, ASF_Data, 0.1)
	ASF_Data2 = data.coeffs_to_ASF(ASF_E, numpy.vstack((ASF_Data,ASF_Data[-1])))
	
	#Test_E = (Output[1:,0]+Output[0:-1,0])*0.5
	#Test_E = numpy.linspace(41257.87,41259.87,num=21)
	#Real_Spectrum2 = KK_PP(Test_E, Output[:,0], Im, Relativistic_Correction)
	
	
	import matplotlib
	matplotlib.use('WXAgg')
	import pylab
	
	pylab.figure()
	pylab.plot(Output[:,0],Output[:,1],'xg-',Output[:,0],Output[:,2],'xb-')
	pylab.plot(ASF_E,ASF_Data2,'+r')
	#pylab.plot(ASF_E,ASF_Data22,'xr')
	pylab.plot(ASF_Data3[0],ASF_Data3[1],'r-')
	#pylab.plot(Test_E,Real_Spectrum2,'*y')
	pylab.xscale('log')
	pylab.show()
	
	
	



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the Kramers-Kronig Calculator software package.
#
# Copyright (c) 2013 Benjamin Watts, Daniel J. Lauk
#
# The software is licensed under the terms of the zlib/libpng license.
# For details see LICENSE.txt

"""Unit tests for module `kk.py`."""

from . import kk
import unittest


class KK_Tests(unittest.TestCase):

	"""Unit tests for the `kk` module."""

	def setUp(self):
		pass

	def tearDown(self):
		pass

	def test_calc_relativistic_correction_known_values(self):
		"""Test `calc_relativistic_correction` for known values."""
		known_values = (
			((1, 1, 1, 1, 1), (1, 1, 1, 1, 1), 4.9998565),
			((1, 2, 3, 4, 5), (1, 1, 1, 1, 1), 14.9973659),
			((1, 2, 3, 4, 5), (5, 4, 3, 2, 1), 34.9952628)
			)
		for Z, stoichiometry, expected in known_values:
			result = kk.calc_relativistic_correction(Z, stoichiometry)
			self.assertTrue(abs(expected - result) < 1E-6,
							"Expected %.6f, got %.6f" % (expected, result))

	def test_coeffs_to_ASF_known_values(self):
		"""Test `coeffs_to_ASF` for known values."""
		known_values = (
			(1, (1, 1, 1, 1, 1), 5),
			(2, (1, 1, 1, 1, 1), 3),
			(3, (1, 2, 3, 4, 5), 6),
			(4, (5, 4, 3, 2, 1), 24)
			)
		for E, coeffs, expected in known_values:
			result = kk.coeffs_to_ASF(E, coeffs)
			self.assertTrue(abs(expected - result) < 1E-6,
							"Expected %.6f, got %.6f" % (expected, result))


if __name__ == "__main__":
	unittest.main()



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kkcalc",
    version="0.8",
    author="Benjamin Watts",
    author_email="benjamin.watts@gmail.com",
    description="A calculator for the Kramers Kronig transform of X-ray spectra",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://bitbucket.org/benajamin/kkcalc",
    packages=setuptools.find_packages(),
    package_data={'kkcalc': ['*.json']},
    install_requires=['numpy', 'scipy','wxPython'],
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: zlib/libpng License  ",
        "Topic :: Scientific/Engineering :: Physics",
        "Operating System :: OS Independent",
    ),
)
